% [if_law,c] = hybifest(x,n,T,gg,A)
%
% This function returns the estimated instantaneous frequency based
% on a hybrid linear/quadratic method. The Gabor decomposition and
% Wigner distribution are used to form the estimate.
% The oversampling of the Gabor decomposition is fixed at Q=4.
%
% See the following paper for details:
% R.G. Baraniuk, M.J. Coates and P. Steeghs, "Hybrid
%  Linear/Quadratic Time-frequency Attributes", to appear
%  Proc. ICASSP 2000.
%
% INPUTS
% x: signal, n: number of time bins in Gabor decomposition, T: time
% spread of the analysis window, gg: Manhattan distance of points
% to include (2-3 generally suitable), A: a pre-generated set of
% vectors for the calculation (see gen_Q4.m)
%
% OUTPUTS
% if_law : the esimated instantaneous frequency
% c      : the Gabor decomposition
% 
% example: [x,iflaw] = fmlin(256,0.05,0.35);
%          [if_est,c] = hybifest(x,32,33,2);
%
% M. Coates, October 1999 - February 2000
%
function [if_law,c] = hybifest(x,n,T,gg,A)

% m number of Gabor coefficients in freq
% n number of Gabor coefficients in time
% n1 signal length

Q = 4;
x = x(:);
n1 = size(x,1);

% Generate Gabor coefficients
c = calc_gab(x,n,Q,T);

% perform some thresholding - avoid unnecessary calculations and smooth
c = c.*(abs(c)>max(max(abs(c)))/10);

m = size(c,1);  % number of frequency bins
Mb=n1/n;        % time-shift between two Gabor coefficients
L = 2*(T-1);    % the window size and size of stored vectors

num = zeros(n1+L,1); % numerator
den = zeros(n1+L,1); % denominator

% either load here or pre-load and pass in, or call the generating function
if (nargin <5)
   if exists(strcat('A4',num2str(n1),'_',num2str(n),'_',num2str(gg),'.mat'))
     load(strcat('A4',num2str(n1),'_',num2str(n),'_',num2str(gg)));
   else
     A = gen_Q4(n,n1,T,d);
   end;
end;

% sigmasq variance of Gaussian window in time domain
sigma_sq = -0.5*((T-1)/2)^2/log(0.005);

tt = (-n1+1):n1;
tt = tt(:);

% A(m,n,m',n')=A(m',n,m',n')=A*(m,n',m',n)=A*(m',n',m,n');
% So treat these four cases together
% When m = m' there are only the two cases A(m,n,m',n')=A*(m,n',m',n)
% When n = n' there are only the two cases (both real) A(m,n,m',n')=A(m',n,m,n')

% Case all four are different - jj, jj_dash frequency indexes;
% ii,ii_dash time indexes
for jj = 2:m
  for jj_dash = max(jj-gg,1):(jj-1)
    for ii = 2:n      
      for ii_dash = max(ii-gg,1):(ii-1)
	if ((jj-jj_dash+ii-ii_dash)<=gg)&(abs(c(jj,ii))>0)&(abs(c(jj,ii_dash))>0)
          % mod operation 
          phase = (jj-jj_dash)*(ii-1+ii_dash-1)-(2*Q)*floor((jj-jj_dash)*(ii-1+ii_dash-1)/(2*Q));
	  ind_s = (Mb*(ii-1+ii_dash-1)/2);
	 
	  v = A{ii-ii_dash+1,jj-jj_dash+1}(phase+1,:).';

          %% Case 1
	  v1 = c(jj,ii)*conj(c(jj_dash,ii_dash))*v;
          s = Mb*(ii-ii_dash)/(2*pi*sigma_sq)*imag(v1);
	  t = ((jj-1)+(jj_dash-1))/(2*m)*real(v1);
	  den((ind_s+1):(ind_s+L)) = den(ind_s+1:ind_s+L) + 2*real(v1);
	  num((ind_s+1):(ind_s+L)) = num(ind_s+1:ind_s+L) + s+t;

          %% Now the reverse case, swapping the time indexes
          v2 = c(jj,ii_dash)*conj(c(jj_dash,ii))*v;
	  s = Mb*(ii_dash-ii)/(2*pi*sigma_sq)*imag(v2);
	  t = ((jj-1)+(jj_dash-1))/(2*m)*real(v2);
	  den((ind_s+1):(ind_s+L)) = den(ind_s+1:ind_s+L) + 2*real(v2);
	  num((ind_s+1):(ind_s+L)) = num(ind_s+1:ind_s+L) + s+t;          
	end;
      end;    
    end;
  end;
end;

% Case when jj == jj_dash, ii~=ii_dash
for jj = 1:m % frequency
  for ii = 2:n % time
    for ii_dash = max(ii-gg,1):(ii-1)
      if (abs(c(jj,ii))>0)&(abs(c(jj,ii_dash))>0)
        v = A{ii-ii_dash+1,1}(1,:);
	v1 = c(jj,ii)*conj(c(jj,ii_dash))*v.';
	s = Mb*(ii-ii_dash)/(2*pi*sigma_sq)*imag(v1);
	t = ((jj-1)+(jj-1))/(2*m)*real(v1);
	ind_s = (Mb*(ii-1+ii_dash-1)/2);
	den((ind_s+1):(ind_s+L)) = den(ind_s+1:ind_s+L) + 2*real(v1);
	num((ind_s+1):(ind_s+L)) = num(ind_s+1:ind_s+L) + s+t;
      end;
    end;
  end;
end;

% Case when jj ~= jj_dash, ii==ii_dash
for jj = 2:m
  for jj_dash = max(jj-gg,1):(jj-1)
    for ii = 1:n
      if ((abs(c(jj,ii))>0)&(abs(c(jj_dash,ii))>0))
        phase = (jj-jj_dash)*(2*ii-2)-(2*Q)*floor((jj-jj_dash)*(2*ii-2)/(2*Q));
        v = A{1,abs(jj-jj_dash)+1}(phase+1,:).';	  
        v1 = c(jj,ii)*conj(c(jj_dash,ii))*v;
	t = ((jj-1)+(jj_dash-1))/(2*m)*real(v1(:));
	ind_s = (Mb*(2*ii-2)/2);
	den((ind_s+1):(ind_s+L)) = den(ind_s+1:ind_s+L)+2*real(v1(:));
	num((ind_s+1):(ind_s+L)) = num(ind_s+1:ind_s+L)+t;
      end;
    end;
  end;
end;

% Case when (jj,ii)==(jj_dash,ii_dash)
for jj = 1:m
  for ii = 1:n
    if (abs(c(jj,ii))>0)
       v = (abs(c(jj,ii))^2)*A{1,1}(1,:);
       ind_s = (Mb*(2*ii-2)/2);
       den(ind_s+1:ind_s+L) = den(ind_s+1:ind_s+L) + v(:);
       num(ind_s+1:ind_s+L) = num(ind_s+1:ind_s+L) + ((jj-1)+(jj-1))/(4*m)*v(:);
    end;
  end;
end;

den2 = den + (den==0);
if_law = num((L/2+1):(n1+L/2))./den2((L/2+1):(n1+L/2));
dv = den((L/2+1):(n1+L/2));
if_law = if_law.*(dv~=0);




